/*
 *  TITLE: IoT Solar Tracker
 *  AUTHOR: Alessandro Canzian
 *  VERSION: 1.9.2
 *  DATE: June 2020
 *  DESCRIPTION: developing an IoT Solar Tracker witn monoaxial rotation
 *              connected to an Online Platform (ThingSpeak) to view in real-time 
 *              all the generated data.
 */


#include "ThingSpeak.h"
#include "WiFiEsp.h"
#include "secrets.h"
#include <Wire.h>
#include <Adafruit_Sensor.h>
#include <Adafruit_BME280.h>

/*-------- upload WiFi Parameters -----------------------------------------*/
char ssid[] = SECRET_SSID;   // your network SSID (name) 
char pass[] = SECRET_PASS;   // your network password
WiFiEspClient  client;
#define ESP_BAUDRATE  115200
unsigned long myChannelNumber = SECRET_CH_ID;
const char * myWriteAPIKey = SECRET_WRITE_APIKEY;
const char * myReadAPIKey = SECRET_READ_APIKEY;
String myStatus = "";
int statusCode=0;


/*-------------------- execution parameters ---------------------------------*/
#define MAXITER 100
#define dataPin 50
#define dirLeft 2
#define dirRight 3
#define speedPinHOR 10
#define input1 A1
#define input2 A2
#define input3 A3
#define input4 A4
#define voltagePin A5
#define currentPin A6

Adafruit_BME280 bme; // I2C

long int readcurrent=0;
long int readvoltage=0;
float dailyenergy=0.0;
long int start_read;
long int stop_read;
long int start_mov;

int lumTOT;
int deltaHOR;

int sNW = 0; 
int sNE = 0; 
int sSW = 0; 
int sSE = 0;
String sensori = "";
int tolHOR = 2;
bool sunlight = false;
bool startup = true;
bool movement = false;
bool stall = false;
bool west = false;
bool east = false;


void setup() 
{
  pinMode(dirLeft, OUTPUT);
  pinMode(dirRight, OUTPUT);
  pinMode(speedPinHOR, OUTPUT);

  //Initialize serial and wait for port to open
  Serial.begin(115200);  // Initialize serial
  
  // initialize serial for ESP module  
  setEspBaudRate(ESP_BAUDRATE);
  
  while (!Serial) {
    ; // wait for serial port to connect. Needed for Leonardo native USB port only
  }

  Serial.print("Searching for ESP8266..."); 
  // initialize ESP module
  WiFi.init(&Serial1);

  // check for the presence of the shield
  if (WiFi.status() == WL_NO_SHIELD) {
    Serial.println("WiFi shield not present");
    // don't continue
    while (true);
  }
  Serial.println("found it!");
   
  ThingSpeak.begin(client);  // Initialize ThingSpeak
  connectWiFi();

// ----------- BME280 check --------------------
  unsigned status;
  status = bme.begin();  
  if (!status) 
  {
      Serial.println("Could not find a valid BME280 sensor, check wiring, address, sensor ID!");
      Serial.print("SensorID was: 0x"); Serial.println(bme.sensorID(),16);
      Serial.print("        ID of 0xFF probably means a bad address, a BMP 180 or BMP 085\n");
      Serial.print("   ID of 0x56-0x58 represents a BMP 280,\n");
      Serial.print("        ID of 0x60 represents a BME 280.\n");
      Serial.print("        ID of 0x61 represents a BME 680.\n");
      while (1);
  }

}


void loop()
{
  int rotspeed = 0; 

  sNW = analogRead(input1);
  sNE = analogRead(input2);
  sSE = analogRead(input3);
  sSW = analogRead(input4); 

  sensori = String(String(sNW) + "; " + String(sSW) + "; " + String(sNE) + "; " + String(sSE) + ".");

  if (movement)
  {
    if ((millis() - start_mov) > 10000)
    {
      Serial.print("stallo secondi: ");
      Serial.println(millis() - start_mov);
      movement = false;
      stall = true;
    }
  }
  
  deltaHOR = (sNW + sSW) - (sNE + sSE);
  lumTOT = sNW + sNE + sSW + sSE;

  if (!movement)
  {
    if (deltaHOR > 2 * tolHOR)
      west = true;
    else if (deltaHOR < - 2 * tolHOR)
      east = true;
  }


	
/*----------  searching dawn ----------------------------------------------------*/  
  if (!sunlight)
  {
    digitalWrite(dirLeft, LOW);
    digitalWrite(dirRight, LOW);

		letturasensori();
		
    if (startup)
    {
      startup = false;
      delay(1000);
    }
    else
      delay(300000);

/*----------  Sun found ----------------------------------------------------*/  
    if(lumTOT > 2400)
    {
      sunlight = true;
      delay(10000);
      start_read = millis();  
      delay(100);
    }
  }
  
/*----------  sunset ----------------------------------------------------*/  
  else if (lumTOT < 1600)
  {
    rotspeed = 200;

    analogWrite(speedPinHOR, rotspeed);
    digitalWrite(dirLeft, HIGH);
    digitalWrite(dirRight, LOW);

    sunlight = false; // light too low

    delay(5000);
  }
  
/*----------  westbound movement ----------------------------------------------------*/  
  else if (west and deltaHOR > -tolHOR and lumTOT > 2800 and !stall)
  {
    if (!movement)
    {
      movement = true;
      start_mov = millis();        
    }

//    rotspeed = (deltaHOR/10) + 150; //delta (0<>2046), rotspeed (40<>255); start from 40 to initiate the motion of the motor
    rotspeed = 200;
    analogWrite(speedPinHOR, rotspeed);
    digitalWrite(dirLeft, LOW);
    digitalWrite(dirRight, HIGH);
  }
  
/*----------  eastbound movement ----------------------------------------------------*/  
  else if (east and deltaHOR < tolHOR and lumTOT > 2800 and !stall)
  {
    if (!movement)
    {
      movement = true;
      start_mov = millis();        
    }

//    rotspeed = -((deltaHOR/10) - 150); //delta (0<>2046), rotspeed (40<>255); start from 40 to initiate the motion of the motor
    rotspeed = 200;
    analogWrite(speedPinHOR, rotspeed);
    digitalWrite(dirLeft, HIGH);
    digitalWrite(dirRight, LOW);
  }
  
/*----------  motor stopped ----------------------------------------------------*/  
  else
  {
    analogWrite(speedPinHOR, rotspeed);
    digitalWrite(dirLeft, LOW);
    digitalWrite(dirRight, LOW);
    
		letturasensori();
    stall = false;
    movement = false;
    east = false;
    west = false;
    
    delay (1000);
  }
   
  delay(2); //wait for X milliseconds
}



/*----------------------------------------------------------------
 * Upload Cloud
 */
void updateCloud(String myStatus, String lum, String temp, String hum, String volt, String curr, String power, String energy, String pressure) 
{
  // set the fields with the values
  ThingSpeak.setField(1, lum);
  ThingSpeak.setField(2, temp);
  ThingSpeak.setField(3, hum);
  ThingSpeak.setField(4, volt);
  ThingSpeak.setField(5, curr);
  ThingSpeak.setField(6, power);
  ThingSpeak.setField(7, energy);
  ThingSpeak.setField(8, pressure);

  // set the status
  ThingSpeak.setStatus(myStatus);
  
  // write to the ThingSpeak channel
  int x = ThingSpeak.writeFields(myChannelNumber, myWriteAPIKey);
  if(x == 200){
    Serial.println("Channel update successful.");
  }
  else{
    Serial.println("Problem updating channel. HTTP error code " + String(x));
  }
  delay(20000);
  
}


/*----------------------------------------------------------------
 * WiFi connection
 */
boolean connectWiFi()
{
  // Connect or reconnect to WiFi
  if(WiFi.status() != WL_CONNECTED)
  {
    Serial.print("Attempting to connect to SSID: ");
    Serial.println(SECRET_SSID);
    while(WiFi.status() != WL_CONNECTED)
    {
      WiFi.begin(ssid, pass);  // Connect to WPA/WPA2 network. Change this line if using open or WEP network
      Serial.print(".");
      delay(5000);     
    } 
    Serial.println("\nConnected.");
  }
  return true;
}



// This function attempts to set the ESP8266 baudrate. Boards with additional hardware serial ports
// can use 115200, otherwise software serial is limited to 19200.

void setEspBaudRate(unsigned long baudrate){
  long rates[6] = {115200,74880,57600,38400,19200,9600};

  Serial.print("Setting ESP8266 baudrate to ");
  Serial.print(baudrate);
  Serial.println("...");

  for(int i = 0; i < 6; i++){
    Serial1.begin(rates[i]);
    delay(100);
    Serial1.print("AT+UART_DEF=");
    Serial1.print(baudrate);
    Serial1.print(",8,1,0,0\r\n");
    delay(100);  
  }
    
  Serial1.begin(baudrate);
}


/**
 *  Reading sensors data and preparing for upload
 */
 
void letturasensori()
{
  float temp = 0.0;
  float humi = 0.0;
  float pres = 0.0;
  float current = 0.0;
  float voltage = 0.0;
  float power = 0.0;
	float tempenergy = 0.0;
  char temp_buffer[25];
  char humi_buffer[25];
  char pres_buffer[25];
  char curr_buffer[25];
  char volt_buffer[25];
  char powe_buffer[25];
  char ener_buffer[25];
	

  readcurrent = 0;
  readvoltage = 0;
  for (int i=0; i<MAXITER; i++)
  {
    readcurrent += analogRead(currentPin);
    readvoltage += analogRead(voltagePin);
    delay(5);
  }
  readcurrent = readcurrent/MAXITER;
  readvoltage = readvoltage/MAXITER;

  Serial.print("corrente:");
  Serial.println(readcurrent);
  Serial.print("tensione:");
  Serial.println(readvoltage);
    
  // 5/1024 -> 4.883mV analog read step
  voltage = (6.3) * ((5 * readvoltage) / 1024.0); // fattore moltiplicativo 6.3 anzichè 6.09 per taratura con dati reali
  current = (5 * (readcurrent - 508) / 1024.0) * (1 / 0.185) ;  // 0 Amp con lettura a 2.5V (508 di lettura analogica); 185mV/A -> 1/0.185 = 5.41A/V
  power = voltage * current;

  temp = bme.readTemperature();
  humi = bme.readHumidity();
  pres = bme.readPressure() / 100.0;

  dtostrf(temp, 2, 2, temp_buffer);
  dtostrf(humi, 2, 2, humi_buffer);
  dtostrf(pres, 2, 2, pres_buffer);
  dtostrf(voltage, 2, 3, volt_buffer);
  dtostrf(current, 2, 3, curr_buffer);
  dtostrf(power, 2, 3, powe_buffer);

// aggiornamento valore energia leggendo l'ultimo valore caricato 
  if(startup)
  {
    myStatus = String("Accensione arduino... " + sensori);
    tempenergy = ThingSpeak.readFloatField(myChannelNumber, 7, myReadAPIKey);  
    statusCode = ThingSpeak.getLastReadStatus();
    if(statusCode == 200)
    {
      Serial.println("Online previous energy: " + String(tempenergy));
      dailyenergy = tempenergy;
    }
    else
    {
      Serial.println("Problem reading channel. HTTP error code " + String(statusCode));
      Serial.println("Using local value of energy: " + String(dailyenergy));
    }    
    delay(100); 
  }
	else if(sunlight)
	{	
		if(stall)
      myStatus = String("Stallo motori " + sensori); 
		else if(current > 2)
			myStatus = String("Produzione ottimale " + sensori); 
		else if(current > 1.5)
			myStatus = String("Buona produzione " + sensori); 
		else if(voltage < 15)
			myStatus = String("Nuvoloso " + sensori);
		else if(current < 1)
			myStatus = String("Produzione bassa " + sensori);
		else
			myStatus = String("Produzione intermedia " + sensori);

		tempenergy = ThingSpeak.readFloatField(myChannelNumber, 7, myReadAPIKey);  
		statusCode = ThingSpeak.getLastReadStatus();
		if(statusCode == 200)
		{
			Serial.println("Online previous energy: " + String(tempenergy));
			if(dailyenergy > tempenergy)
      {  
        stop_read = millis();
        float interval_sec = (stop_read - start_read) / 1000.0;
        dailyenergy += power * interval_sec/3600.0;
        start_read = millis();  
      }
      else
      {
        dailyenergy = tempenergy;
        stop_read = millis();
  			float interval_sec = (stop_read - start_read) / 1000.0;
  			dailyenergy += power * interval_sec/3600.0;
  			start_read = millis();  
      }
		}
		else
		{
			Serial.println("Problem reading channel. HTTP error code " + String(statusCode));
			Serial.println("Using local value of energy: " + String(dailyenergy));
			stop_read = millis();
			float interval_sec = (stop_read - start_read) / 1000.0;
			dailyenergy += power * interval_sec/3600.0;
			start_read = millis();  
		}    
	}
  else
  {
    myStatus = String("Periodo notturno, ricerca alba... " + sensori);
    dailyenergy = 0;
  }    
	dtostrf(dailyenergy, 2, 3, ener_buffer);

  if(connectWiFi())
    updateCloud(myStatus, String(lumTOT), String(temp_buffer), String(humi_buffer), String(volt_buffer), String(curr_buffer), String(powe_buffer), String(ener_buffer), String(pres_buffer));
}
